
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'compat/kernel/constant'
require 'eventmachine'

module Tomte

  autoload :Actor,        'tomte/async/actor'
  autoload :Worker,       'tomte/async/worker'
  autoload :Director,     'tomte/async/director'

  module Workers; end

module Agent

  autoload :Common,       'tomte/agent/common'

  #
  # This is the async agent, the workhorse of the Tomte framework
  #
  #
  class Async < Tomte::Agent::Common

    include Tomte::Agent::Daemon

    # this is the director actor
    attr_accessor :director

    #
    # gives you an Async Agent
    #
    # no options so far
    #
    def initialize(options = {})
      super

    end

    #
    # actually runs the agent
    # - daemonize, if requested
    # - runs the director worker
    # - runs the other system workers
    # - runs configured non-system workers
    #
    def run!(options = {})
      super

    end

    #
    # shuts down this agent
    #
    #
    def shutdown!(options = {})

      if EM.reactor_running? && director
        EM.next_tick { director.shutdown }
      else
        daemon_shutdown
        super
      end
    end

    def reactor_thread
      EM.reactor_thread
    end

    private
    #######

    #
    # called when the director has done shutting down and
    # we can continue closing the agent
    #
    def deferred_shutdown
      log.debug "- Deferred agent shutdown."

      daemon_shutdown
      after_shutdown
    end

    #
    # this is where the async agent really starts
    #
    def start_agent(options = {})

      Thread.abort_on_exception = true
      initialize_director(options)

      #trap('TERM') { shutdown! }
      #trap('INT') { shutdown! }

      setup_signal_traps

      # the original thread does nothing so far, just runs an idle cycle
      # it's just left for a future implementation
      run

    end


    #
    # the director thread is an EM reactor loop doing only one thing, running the director.
    #
    #
    #
    def initialize_director(options)

      #
      # The director thread is inside a EM reactor loop
      #
      ensure_reactor_running!
      EM.schedule do
        @director = Tomte::Director.new(config.merge( :logger => @log,
                                                      :configuration => config,
                                                      :after_shutdown => method(:deferred_shutdown)))
      end
    end

    def ensure_reactor_running!

      unless EM.reactor_running?

        Thread.abort_on_exception = true
        log.debug "+ Async agent is starting the reactor."

        thr = Thread.current
        Thread.new { EM.run { EM.next_tick { thr.wakeup } } }
        Thread.stop
        log.debug "+ Async agent has done starting the reactor."
      end

      EM.error_handler { |e| stray_exception e }
    end

    #
    # void loop, do nothing but checking for pending signals
    #
    def run
      loop do
        handle_signals
        sleep 1
      end
    end


  end
end
end
