
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'pp'
require 'tomte/structures/registry'
require 'tomte/async/supervisor/database_registry_store'

module Tomte
module Workers
  class Supervisor < Tomte::Worker
    class AgentRegistry
      extend Forwardable

      DEFAULT_MISSED_TICKS = 2

      MANDATORY_IE = [:identity, :host, :uuid].freeze


      attr_accessor :log
      attr_accessor :registry
      # how many heartbeats to miss before agent goes inactive
      attr_accessor :missed_ticks
      # how long an inactive agent stays in the inactive list (in seconds)
      attr_accessor :inactive_time

      def_delegators :@registry, :[], :has_key?

      def initialize(log, heartbeat)

        @heartbeat = heartbeat
        @missed_ticks = DEFAULT_MISSED_TICKS
        @inactive_time = @heartbeat * DEFAULT_MISSED_TICKS
        @log = log
        @registry = Tomte::Structures::Registry.new(DatabaseRegistryStore)

        registry_sync
      end

      #
      # check if a MM has all the requested headers
      #
      def check_header_ie(header)
        MANDATORY_IE.each { |i|  return "missing ie: '#{i}'" unless header.has_key? i }
        false
      end

      #
      # check if a message is well formed
      #
      def check_message(message)
        begin
          res = check_header_ie(message.header)
          return res if res
        rescue
          return 'malformed message'
        end
        false
      end



      #
      # learns a new agent
      #
      def learn(id, header, info, time = Time.now.to_i)

        raise ArgumentError, "missing mandatory IE" if check_header_ie(header)
        log.debug("- #{id} is a new entry in registry.")

        if info.capabilities.include? :hb
          registry.learn(id, :active) do |a|
            a.state = :active
            a.signed_on = time
            a.hb_time = info.hb_time || @heartbeat
            a.interval = a.hb_time * @missed_ticks
            a.capabilities = info.capabilities
            a.info = info
            copy_ie info.local, a
          end
          touch(id, time)
        else
          registry.learn(id, :static) do |a|
            a.state = :static
            a.signed_on = time
            a.capabilities = info.capabilities
            a.since = time
            a.info = info
            copy_ie info.local, a
          end
        end
      end

      #
      # forget an agent by id
      #
      def forget(id)
        log.debug("- #{id} is deleted")
        registry.forget(id)
      end
      alias :delete  :forget

      #
      # touch the agent in the registry
      #
      def touch(id, time = Time.now.to_i)

        return unless registry.has_key? id

        if(registry[id].state = :active)
          registry[id].expires = time + registry[id].interval
          registry.touch(id, :active, registry[id].expires)
          registry[id].last_seen = time
          registry[id].since = time
        else
          go_active(id)
        end
        registry.commit id
      end

      #
      # mark the agent as active
      #
      def go_active(id, time = Time.now.to_i)
        log.debug("- #{id} is active.")
        registry[id].expires = time + registry[id].interval
        registry.move(id, :active, registry[id].expires)
        registry[id].state = :active
        registry[id].last_seen = time
        registry[id].since = time
        registry.commit id
      end


      #
      # mark the agent as inactive
      #
      def go_inactive(id, time = Time.now.to_i)
        log.debug("- #{id} is inactive.")
        registry[id].expires = time + @inactive_time
        registry.move(id, :inactive, registry[id].expires)
        registry[id].state = :inactive
        registry[id].since = time
        registry.commit id
      end


      #
      # mark the agent as dead
      #
      def go_dead(id, time = Time.now.to_i)
        log.debug("- #{id} is dead.")
        registry.move(id, :dead)
        registry[id].state = :dead
        registry[id].since = time
        registry.commit id
      end

      #
      # expire all dead elements
      #
      def flush
        registry.flush :dead
      end

      #
      # ran every tick to move agent in registry
      #
      def check_expire(time)

        registry.expire(:active, time) do |id, info|
          go_inactive id
        end

        registry.expire(:inactive, time) do |id, info|
          go_dead id
        end

      end

      #
      # called from timer
      #
      def tick(time = Time.now_to_i)
        check_expire(time)
      end

      #
      # dumps the registry for transmitting remotely
      #
      def dump
        ## TODO
      end

      #
      # loads registry from serialized data
      #
      def load(data)
        ## TODO
      end

      #
      # syncronizes the queues with the saved states
      #
      def registry_sync
        @registry.each do |k, info|
          if k.state
            time = info.last_seen.to_i +
            registry.move(id, k.state, time)
          end
        end
      end

      #
      # clears the registry
      #
      def clear
        registry.clear
      end
      alias :reset! :clear

      def to_h
        registry.to_h
      end

      def inspect
        "#{self.class.name} ----------\n" + registry.inspect + "\n---\n"
      end

      def to_human
        out  = "-- registry --\n"
        registry.info.each do |id, info|
          out += "== #{id}\n"
          info.each do |key, value|
            out += "|  %15s: %s\n" % [key, value]
          end
        end
        out
      end
      alias :show :to_human


      private
      #######

      def copy_ie(header, info)
        MANDATORY_IE.each { |ie| info[ie] = header.send(ie)}
      end

    end
  end
end
end

