# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Network
  module API

    #
    # the blankslate class that you extend to implement an API
    #
    class Server

      class_attribute :defined_rpc_methods
      self.defined_rpc_methods = {}

      class Method
        Options = Struct.new(
            :raw,
            :deferred,
            :human_readable,
            :help_caption,
            :help_text
        ) do
            def to_h
              Hash[members.zip values]
            end
        end

        attr_reader :verb
        attr_reader :opts

        def initialize(verb, options = {})
          @verb = verb
          @opts = Options.new

          opts.raw =            options.fetch(:raw,            false)
          opts.deferred =       options.fetch(:deferred,       false)
          opts.human_readable = options.fetch(:human_readable, false)
          opts.help_caption =   options.fetch(:help_caption,   verb.to_s)
          opts.help_text =      options.fetch(:help_text,      '- undocumented method -')
        end


        def help_caption(text)
          opts.help_caption = text
        end

        def help_text(text)
          opts.help_text = text
        end

        def deferred_method
          opts.deferred = true
        end

        def human_readable
          opts.human_readable = true
        end

        def raw
          opts.raw = true
        end
      end


      attr_reader :log

      #
      # the name for this API
      #
      attr_reader :api_name

      #
      # the worker that started this api
      #
      attr_reader :worker

      #
      # the (optional) target of this API, defaults to worker, if not given
      #
      attr_reader :target


      def initialize(worker, params = {})
        @api_name   = params.fetch(:name, 'unnamed-api')

        @worker = worker
        @target = params.fetch(:target, worker)

        @log = @worker.log

        log.debug "- #{self.class}: New API server."
      end

      #
      # called when the server has completed initializing
      #
      def _api_initialize(server)
        log.debug "- #{self.class}: API initialization".yellow.bold

        server.allowed_methods = defined_rpc_methods.map do | v, m |
          log.warn "- #{self.class} defines unimplemented RPC method '#{m.verb}'." unless respond_to?(m.verb)
          m.verb
        end
        server.raw_methods = self.class.raw_rpc_methods.map {|v, m| m.verb }
        server.deferred_methods = self.class.deferred_rpc_methods.map {|v, m| m.verb }

        log.debug "- Allowed methods: " + server.allowed_methods.inspect.blue
        log.debug "- Raw methods: " + server.raw_methods.inspect.blue
        log.debug "- Deferred methods: " + server.deferred_methods.inspect.blue
      end

      #
      # called when the server is connected
      #
      def _api_connect(server)
        log.debug "- #{self.class}: API initialization"
      end

      def _api_before_request(server, request)
        Thread.current[:current_request] = request
      end

      def _api_after_request(server, request)

      end

      def current_request
        Thread.current[:current_request]
      end

      def human_output(data = '', &block)
        return unless current_request.payload.wants_human_output?
        current_request.response.payload.human_output ||= ''
        current_request.response.payload.human_output += data + (block ? block.call.to_s : '')
      end

      private
      #######

      class << self

        def inherited(subclass)
          subclass.defined_rpc_methods = defined_rpc_methods.dup
        end

        def raw_rpc_methods
          defined_rpc_methods.select { |v, m| m.opts.raw }
        end

        def deferred_rpc_methods
          defined_rpc_methods.select { |v, m| m.opts.deferred }
        end


        def rpc_methods(*verbs, &proc)
          verbs.each { |v| rpc_method(v, {}, &proc) }
        end

        def rpc_method(verb, options = {}, &proc)
          m = Method.new(verb, options)
          m.instance_exec(&proc) if proc
          defined_rpc_methods[verb.to_s] = m
          m
        end
      end


      def banner_versions

        { :engine    => RUBY_ENGINE,
          :runtime   => RUBY_VERSION,
          :framework => Tomte::VERSION,
          :protocol  => Tomte::Protocol::VERSION,
          :agents    => Tomte::Agents::VERSION,
          :api       => _api_version }
      end

      def _api_version
        "0.0.0"
      end

      #
      ##### PUBLIC RPC METHODS ###############################################

      public
      ######
      #
      # predefined methods
      #

      rpc_method :banner do
        help_caption 'Heading information for this API'
      end

      def banner
        human_output do
          <<-EOT.unindent
          == #{api_name} == #{self.class.name} ==
          Version: #{_api_version}

          EOT
        end

        {:api => api_name.to_s, :class => self.class.name, :versions => banner_versions}
      end


      rpc_method :api_help

      def api_help(*args)
        human_output do

          case args.first
          when nil
            "== HELP ==".blue.bold
          when String, Symbol
            m = args.first.to_s
            if defined_rpc_methods.has_key?(m)
              method = defined_rpc_methods.fetch(m)
              sprintf("--- %s ---\n%s\n\n%s", method.verb.to_s.blue.bold,
                     method.opts.help_caption.yellow, method.opts.help_text )
            else
              " -- no help for method '#{m}'.".blue.bold
            end

          end
        end
      end

      rpc_method :ping

      def ping
        :pong
      end

      rpc_method :interface do
        help_caption 'Displays the service banner for the API'
      end


      def interface
        human_output do
          out = "    == #{api_name} == #{self.class.name} ==\n".blue.bold
          out << defined_rpc_methods.values.sort{ |x,y| x.verb <=> y.verb }.map do |data|
                    sprintf("%30s: %s", data.verb.to_s.yellow, data.opts.help_caption)

                  end.join("\n")
        end
        defined_rpc_methods.inject_hash { |h, (v,m)| h[v] = m.opts.to_h }
      end


      def version

      end


    end
  end
end
