
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

# Simple dispatcher.
#
# Provides optional agents logon/logoff, periodic updates on the workers status and wall services


module Tomte::Workers
  class SimpleDispatcher < Tomte::Worker

    module WatchdogHandler

      def post_init
        send_data "200 watchdog\n"
      end

      def receive_data data
        send_data data
        close_connection if data =~ /quit/i
      end

      def unbind
      end
    end

    MANAGEMENT_EVENTS_EXCHANGE = 'tomte.mgmt.events'
    WALL_EXCHANGE = 'tomte.mgmt.wall'

    DEFAULT_HELLO_INTERVAL = 30

    attr_accessor :wall_exchange
    attr_accessor :management_events_exchange

    worker :system => true, :do_not_registry => true

    bus :management do |ep|
      @management_events_exchange = ep.producer(:exchange => @management_events_exchange_name,
                                                :exchange_type => :fanout)
      @wall_exchange = ep.producer(:exchange => @wall_exchange_name, :exchange_type => :fanout)
    end

    def initialize(director=nil, params = {})
      super
      @management_events_exchange_name =
                            config.agents.fetch(:management_events_exchange, MANAGEMENT_EVENTS_EXCHANGE)
      @wall_exchange_name = config.agents.fetch(:wall_exchange,              WALL_EXCHANGE)
      @hello_interval =     config.agents.fetch(:hello_interval,             DEFAULT_HELLO_INTERVAL)
      @hello_enabled =      config.agents.fetch(:hello_enabled,              false)
      @watchdog_socket =    config.agents.fetch(:watchdog_socket,            nil)
    end

    def init(options = {})
      super
      log.warn "= Using the simple dispatcher. Supervision is one-way.".blue

      start_watchdog_server
      setup_event_exchanges do
        heartbeat_start
        director << :dispatcher_ready
      end
    end

    def events
      receive do |filter|

        #
        # sign on the network
        #
        filter.when :sign_on do
          sign_on_network do
            director << :network_ready
          end
        end

        #
        # sign off the network
        #
        filter.when :shutdown do
          log.debug "- Dispatcher been requested to sign-off the network and shut down."

          append_callback :after_shutdown do
            director << :dispatcher_exited
          end

          sign_off_network do
            shutdown
          end
        end

        #
        # send an update message
        #
        filter.when :workers_update do
          send_network_update
        end

        #
        # a new worker has been started
        #
        filter.when(T[:new_worker, Tomte::Worker]) do |_, worker|
          send_worker_start worker
        end


        #
        # a wall message
        #
        filter.when(T[:wall, Symbol, String, Hash]) do |_, severity, caption, attributes|
          publish_wall_message severity, caption, attributes
        end

      end
    end

    private


    def start_watchdog_server
      if(@watchdog_socket)
        log.debug "- Dispatcher, starting watchdog server on #{@watchdog_socket}."
        @watchdog = EventMachine.start_unix_domain_server(@watchdog_socket, WatchdogHandler)
      end
    end

    def stop_watchdog_server
      if @watchdog
        EventMachine.stop_server(@watchdog) if @watchdog
        FileUtils.rm(@watchdog_socket, :force => true)
      end
    end

    def setup_event_exchanges(&when_done)
      @management_events_exchange.connect! { @wall_exchange.connect! { when_done.call }}
    end

    def heartbeat_start
      return unless @hello_enabled

      #
      # should send an agent presence message
      #
      @hello_timer = Tomte::Timer.periodic(@hello_interval) do
        send_hello
      end
    end

    def sign_on_network(&when_done)
      log.debug "- Dispatcher is signing-on to the network.".yellow
      management_events_exchange.publish(
        Tomte::Network::Mgmt::Events::SignOn.new(
            :hello_enabled => @hello_enabled,
            :hello_interval => @hello_interval
        ),
        &when_done
      )
    end

    def sign_off_network(&when_done)
      return(when_done.call) if !management_events_exchange

      log.debug "- Dispatcher is signing off the network.".yellow
      begin
        management_events_exchange.publish(
          Tomte::Network::Mgmt::Events::SignOff.new(:reason => 'Signing Off.'),
          errback: -> {
            log.error " = sign_off_network failed".red
            when_done.call
          },
          &when_done
        )
      rescue 
        log.error " |  #{$!.to_s}".red
        log.error " |   #{$!.backtrace.join("\n |   ")}".blue
        #already closed?
        when_done.call
      end
    end

    def send_hello
      send_network_update
    end

    def send_network_update
      log.debug "- Dispatcher: sending network update."
      management_events_exchange.publish(
          Tomte::Network::Mgmt::Events::Update.new(
              :workers => director.registry.workers_list.map do |w|
                Tomte::Network::Mgmt::Worker.new w
              end
          ),
          :expiration => @hello_interval * 1000
      )
    end

    def send_worker_start(worker)
      send_network_update
    end

    def shutdown
      stop_watchdog_server
      super
    end

    def publish_wall_message(severity, caption, attributes)
      log.debug "- sending wall message: "
      log.debug Hash.new(:severity => severity, :caption => caption, :attributes => attributes).ai

      expiration = attributes.delete(:expiration)

      message = Tomte::Network::Mgmt::Wall::Message.new(attributes)
      message.severity = severity
      message.caption = caption

      wall_exchange.publish(message, :expiration => expiration)
    end
  end
end
